(*
簡易HTMLパーサ
XMLとして破綻しているHTMLをXMLパーサを通せる程度に構造化する。
入力はUTF-8を想定
OCaml 4.02
2015.02.11 - 2015.04.04 Takeshi Yoneki
ostra@asahi.email.ne.jp
*)

open Tools

exception ERROR_PARSE
exception ERROR_BUG

type t = 
  | Element of (string * (string * string) list * t list)
  | PCData of string
  | CData of string
  | Comment of string
  | Empty

type partKind = PK_Start | PK_End | PK_Empty | PK_Text | PK_Comment | PK_Doctype | PK_Decl | PK_CData

module StringSet = Set.Make(String)

let parsePhase1 data =
  let part = ref ([]: char list) in
  let pool = ref ([]: (partKind * string) list) in
  let inTag = ref false in
  let inQuote = ref false in
  let inComment = ref false in
  let inCData = ref false in
  let addPartToPool () =
    if not (List.is_empty !part) then
      let row = CharList.to_string_rev !part in
      let s = String.lowercase row in
      let k = 
        if String.starts_with s "<!doctype" && String.ends_with s ">" then
          PK_Doctype
        else if String.starts_with s "<?" && String.ends_with s "?>" then
          PK_Decl
        else if String.starts_with s "<![cdata[" && String.ends_with s "]]>" then
          PK_CData
        else if String.starts_with s "<!--" && String.ends_with s "-->" then
          PK_Comment
        else if String.starts_with s "</" && String.ends_with s ">" then
          PK_End
        else if String.starts_with s "<" && String.ends_with s "/>" then
          PK_Empty
        else if String.starts_with s "<" && String.ends_with s ">" then
          PK_Start
        else
          PK_Text
      in
      let _ = pool := (k, row) :: !pool in
      part := []
    else
      ()
  in
  let _ =
    String.iter
      (fun ch ->
        if ch = '\r' then
          ()
        else if !inCData then
          let _ = part := ch :: !part in
          if ch = '>' && String.ends_with (CharList.to_string_rev !part) "]]>" then
            let _ = addPartToPool () in
            inCData := false
          else
            ()
        else if !inComment then
          let _ = part := ch :: !part in
          if ch = '>' && String.ends_with (CharList.to_string_rev !part) "-->" then
            let _ = addPartToPool () in
            inComment := false
          else
            ()
        else if !inQuote then
          let _ = 
            if ch = '\"' then
              inQuote := false
            else
              ()
          in
          part := ch :: !part
        else
          let _ =
            if ch = '>' then
              let _ = part := ch :: !part in
              let _ = addPartToPool () in
              inTag := false
            else
              let _ =
                if ch = '\"' then
                  inQuote := !inTag
                else if ch = '<' then
                  let _ = addPartToPool () in
                  inTag := true
                else
                  ()
              in
              part := ch :: !part
          in
          if ch = '-' && (CharList.to_string_rev !part) = "<!--" then
              inComment := true
          else if ch = '[' && String.lowercase (CharList.to_string_rev !part) = "<![cdata[" then
              inCData := true
          else
            ()
      )
      data
  in
  List.rev !pool

let fixCharEntity tx =
  (* 実体参照は文字にする *)
  let rec loop t ac =
    let (n, p) = CharEntity.find_char_entity t in
    if n < 0 then
      String.concat "" (List.rev (t :: ac))
    else
      let tl = String.length t in
      let pl = String.length p in
      let ac = (CharEntity.decode_char_entity p) :: (String.sub t 0 n) :: ac in
      let tr = String.sub t (n + pl) (tl - (n + pl)) in
      loop tr ac
  in
  loop tx []

let parseStart k row =
  let part = ref ([]: char list) in
  let pool = ref ([]: string list) in
  let inQuote = ref false in
  let sub =
    if k = PK_Empty then
      (String.sub row 1 ((String.length row) - 3)) ^ " "
    else
      (String.sub row 1 ((String.length row) - 2)) ^ " "
  in
  let isSpace ch =
    ch = ' ' || ch = '\t' || ch = '\n' || ch = '\r'
  in
  let _ =
    String.iter
      (fun ch ->
        let enter n () =
          if List.length !part > 0 then
            let row = CharList.to_string_rev !part in
            let _ = pool := row :: !pool in
            part := []
          else
            ()
        in
        if !inQuote then
          let _ = 
            if ch = '\"' then
              let _ = enter 1 () in
              inQuote := false
            else
              part := ch :: !part
          in
          ()
        else
          let _ =
            if ch = '\"' then
              let _ = enter 2 () in
              inQuote := true
            else if isSpace ch || ch = '=' then
              enter 3 ()
            else
              part := ch :: !part
          in
          ()
      )
      sub
  in
  let px = List.rev !pool in
  let tag = List.hd px in
  let px = List.tl px in
  let px =
    if (List.length px) mod 2 <> 0 then
      List.take ((List.length px) - 1) px
    else
      px
  in
  let ax = List.divide 2 px in
  let attr =
    List.map
      (fun pair ->
        let n = List.hd pair in
        let v = List.hd (List.tl pair) in
        (n, fixCharEntity v)
      )
      ax
  in
  Element (tag, attr, [])
  
let parseEnd row =
  Element (String.trim (String.sub row 2 ((String.length row) - 3)), [], [])

let fixComment t =
  (* コメント内に"--"を記載してしまうページがある *)
  let s = String.sub t 4 ((String.length t) - 7) in
  let s = String.replace s "--" "&#x2d;&#x2d;" in
  s

let parsePhase2 data =
  List.map
    (fun (k, row) ->
      let part =
        if k = PK_Start || k = PK_Empty then
          parseStart k row
        else if k = PK_End then
          parseEnd row
        else if k = PK_Text then
          PCData (fixCharEntity row)
        else if k = PK_CData then
          CData (String.sub row 9 ((String.length row) - 12))
        else if k = PK_Comment then
          Comment (fixComment row)
        else
          PCData (row)
      in
      (k, part)
    )
    data

(*let print_stack ls =
  print_endline (String.concat " " ls)*)

(* 末尾再帰にできそうにないのでループにする *)
let parsePhase3 data =
  let endTags =
    StringSet.of_list
      (List.map
        (fun (_, p) -> 
          match p with
          | Element (tag, _, _) ->
            String.lowercase tag
          | _ -> 
            raise ERROR_BUG
        )
        (List.filter (fun (k, _) -> k = PK_End) data))
  in
  (*let _ = StringSet.iter (fun s -> print_endline s) endTags in*)
  let hasEnd tag =
    StringSet.mem (String.lowercase tag) endTags
  in
  let stack = ref [""] in
  let dx = ref data in
  let rec mkTree () =
    let break = ref false in
    let sibling = ref [] in
    let q = ref "" in
    let _ =
      while !dx <> [] && not !break do
        let (k, p) = List.hd !dx in
        let _ = dx := List.tl !dx in
        if k = PK_Start then
          match p with
          | Element (tag, attr, _) ->
            if hasEnd tag then
              let _ = stack := tag :: !stack in
              let (child, cue) = mkTree () in
              let _ = stack := List.tl !stack in
              let e = Element (tag, attr, List.rev child) in
              let _ = sibling := e :: !sibling in
              if List.length !stack > 1 then
                if cue <> tag && List.mem cue !stack then
                  let _ = q := cue in
                  break := true
                else
                  ()
              else
                ()
            else
              sibling := p :: !sibling
          | _ -> 
            raise ERROR_BUG
        else if k = PK_End then
          match p with
          | Element (tag, _, _) ->
            let _ = q := tag in
            break := true
          | _ -> 
            raise ERROR_BUG
        else if k = PK_Text || k = PK_CData || k = PK_Empty || k = PK_Comment then
          sibling := p :: !sibling
        else
          (* DOCTYPEなどは捨てる *)
          ()
      done
    in
    (!sibling, !q)
  in
  let (s, q) = mkTree () in
  s

let print_htmlp p =
  let rec subfn p sp =
    match p with
    | Element (tag, attr, child) ->
      let _ = print_endline (sp ^ tag) in
      let _ =
        List.iter
          (fun (n, v) ->
            print_endline (sp ^ n ^ "=" ^ v)
          )
          attr
      in
      let _ =
        List.iter
          (fun c ->
            subfn c (sp ^ " ")
          )
          child
      in
      ()
    | PCData (text) ->
      let t = String.trim text in
      if not (String.is_empty t) then
        print_endline (sp ^ t)
      else
        ()
    | _ ->
      ()
  in
  subfn p ""

(*
let print_phase2 p =
  List.iter
    (fun (k, part) ->
      print_htmlp part
    )
    p

let print_phase3 p =
  List.iter
    (fun e ->
      print_htmlp e ""
    )
    p
*)

let parse data =
  let p1 = parsePhase1 data in
  let p2 = parsePhase2 p1 in
  (*let _ = print_phase2 p2 in*)
  let p3 = parsePhase3 p2 in
  (*let _= print_phase3 p3 in*)
  let p4 = 
    List.filter
      (fun e ->
        match e with
        | Element (_, _, _) ->
          true
        | _ ->
          false
      )
      p3
  in
  if List.is_empty p4 then
    Empty
  else
    List.hd p4

let to_xml_string elm =
  (* 実体参照にすべき文字は実体参照にする *)
  let toCharEntity isAttr t =
    let t = String.replace t "&" "&amp;" in
    let t = String.replace t ">" "&gt;" in
    let t = String.replace t "<" "&lt;" in
    let t = 
      if isAttr then
        String.replace t "\"" "&quot;"
      else
        t
    in
    t    
  in
  let fixAttrValue t =
    toCharEntity true t
  in
  let fixText t =
    toCharEntity false t
  in
  let rec onNode e ac =
    match e with
    | Element (tag, attr, child) ->
      let axs = 
        List.map
          (fun (n, v) ->
            n ^ "=\"" ^ (fixAttrValue v) ^ "\""
          )
          attr
      in
      let ax = String.concat " " axs in
      let s = "<" ^ tag ^ 
        if String.is_empty ax then
          ""
        else 
          " " ^ ax
      in
      if List.is_empty child then
        (s ^ " />") :: ac
      else
        let ac = (s ^ ">") :: ac in
        let ac = 
          List.fold_left
            (fun ac e ->
              onNode e ac
            )
            ac
            child
        in
        ("</" ^ tag ^ ">") :: ac
    | Comment (t) ->
      ("<!--" ^ t ^ "-->") :: ac
    | PCData (t) ->
      (fixText t) :: ac
    | CData (t) ->
      (fixText t) :: ac
    | _ ->
      ac
  in
  String.concat "" (List.rev (onNode elm []))

let is_element node =
  match node with
  | Element _ -> true
  | _ -> false

let is_text node =
  match node with
  | PCData _ -> true
  | CData _ -> true
  | _ -> false

(*
let tag node =
  match node with
  | Element (t, _, _) -> t
  | _ -> ""

let attribs node =
  match node with
  | Element (_, a, _) -> a
  | _ -> []

let children node =
  match node with
  | Element (_, _, c) -> c
  | _ -> []
*)

let set_attribute elm name value =
  match elm with
  | Element (t, a, c) ->
    let att = List.filter (fun (n, v) -> n <> name) a in
    let ax = (name, value) :: att in
    Element (t, ax, c)
  | _ ->
    elm  

let remove_attribute elm name =
  match elm with
  | Element (t, a, c) ->
    let att = List.filter (fun (n, v) -> n <> name) a in
    Element (t, att, c)
  | _ ->
    elm  

let get_attribute elm name =
  match elm with
  | Element (_, a, _) ->
    let att = List.filter (fun (n, v) -> n = name) a in
    if List.is_empty att then
      ""
    else
      let (_, v) = List.hd att in
      v
  | _ -> 
    ""

let get_element_text elm =
  match elm with
  | Element (_, _, c) ->
    let ts =
      List.filter_map
        (fun x ->
          match x with
          | PCData t -> Some t
          | CData t -> Some t
          | _ -> None
        )
        c
    in
    let tx = String.concat "" ts in
    tx
  | _ -> 
    ""

let get_elements_by_tag_name elm name =
  let nsname = ":" ^ name in
  let rec subfn elm =
    match elm with
    | Element (t, a, c) ->
      let r1 =
        List.filter_map
          (fun n ->
            if is_element n then
              Some (subfn n)
            else
              None
          )
          c
      in
      let r2 = List.concat r1 in
      let r3 =
        if t = name || String.ends_with t nsname then
          elm :: r2
        else
          r2
      in
      r3
    | _ ->
      []
  in
  subfn elm

(*
let get_element_by_tag_name elm name =
  let r = get_elements_by_tag_name elm name in
  if List.length r = 0 then
    None
  else
    Some (List.hd r)
*)
