(*
http.ml
module Http
当面目的はHTTP GET
OCaml 4.02
2015.03.09 Takeshi Yoneki
ostra@asahi.email.ne.jp
*)

open Tools

let buffer_size = 4096

module Dict = Map.Make(String)

type t = {
  filedesc: Unix.file_descr;
  buffer: string ref;
  readsize: int ref;
  point: int ref;
  eof: bool ref;
  status: string;
  resp_header: string Dict.t
}

let parse_url url =
  let r = String.split ["/"] url in
  if List.length r < 2 then
    ("", "", 0, "")
  else
    let s = List.hd r in
    let r = List.tl r in
    let protocol = String.sub s 0 ((String.length s) - 1) in
    let s = List.hd r in
    let r = List.tl r in
    let h = String.split [":"] s in
    let (host, port) = 
      if List.length h > 1 then
        (List.hd h, List.hd (List.tl h))
      else
        let p = 
          match protocol with
          | "http" -> "80"
          | "https" -> "443"
          | _ -> "0"
        in
        (s, p)
    in
    let path = 
      if not (List.is_empty r) then
        String.concat "/" r
      else
        ""
    in
    let path =
      if String.ends_with url "/" then
        path ^ "/"
      else
        path
    in
    let path = "/" ^ path in
    (protocol, host, int_of_string port, path)

let getc fp =
  let _ = 
    if !(fp.eof) then
      ()
    else 
      if !(fp.readsize) <= !(fp.point) then
        let _ = fp.readsize := Unix.recv fp.filedesc !(fp.buffer) 0 buffer_size [] in
        (*let _ = print_endline (string_of_int !(fp.readsize)) in*)
        let _ = fp.point := 0 in
        if !(fp.readsize) = 0 then
          fp.eof := true
        else
          ()
      else
        ()
  in
  if !(fp.eof) then
    None
  else if !(fp.readsize) < buffer_size && !(fp.readsize) <= !(fp.point) then
    None
  else
    let c = !(fp.buffer).[!(fp.point)] in
    let _ = fp.point := !(fp.point) + 1 in
    Some c

let read fp len =
  let rec loop ac i =
    if len > 0 && i >= len then
      ac
    else
      match getc fp with
      | None -> 
        ac
      | Some c -> 
        loop (c :: ac) (i + 1)
  in
  CharList.to_string_rev (loop [] 0)

let resp_header fp name =
  let n = String.lowercase name in
  if Dict.mem n fp.resp_header then
    Dict.find n fp.resp_header
  else
    ""

let resp_header_keys fp =
  Dict.fold (fun k _ ls -> k :: ls) fp.resp_header []

let close fp =
  Unix.shutdown fp.filedesc Unix.SHUTDOWN_ALL

let rec open_get req_header url =
  (*let _ = print_endline url in*)
  let (protocol, host, port, path) = parse_url url in
  let h = Unix.gethostbyname host in
  let fd = Unix.socket h.Unix.h_addrtype Unix.SOCK_STREAM 0 in
  (*let _ = print_endline h.Unix.h_addr_list.(0) in*)
  let _ = Unix.connect fd (Unix.ADDR_INET (h.Unix.h_addr_list.(0), port)) in
  let rh = 
    if List.length req_header > 0 then
      let rhs = List.map (fun (n, v) -> n ^ ": " ^ v) req_header in
      (String.concat "\r\n" rhs) ^ "\r\n"
    else
      ""
  in  
  let msg = "GET " ^ path ^ " HTTP/1.1\r\nHost: " ^ host ^ "\r\n" ^ rh ^ "\r\n" in
  (*let _ = print_endline msg in*)
  let _ = Unix.send fd msg 0 (String.length msg) [] in
  let fp = { 
    filedesc = fd; 
    buffer = ref (Bytes.create buffer_size);
    readsize = ref 0;
    point = ref 0;
    eof = ref false;
    status = "";
    resp_header = Dict.empty
  }  
  in  
  let rec rdHead st m line =
    match getc fp with
    | None ->
      (st, m)
    | Some c ->
      if c = '\n' then
        if String.is_empty line then
          (st, m)
        else
          if String.is_empty st then
            let lx = String.split [" "] line in
            if List.length lx >= 2 then
              rdHead (List.hd (List.tl lx)) m ""
            else
              rdHead st m ""
          else
            let lx = String.split_n [":"] 2 line in
            if List.length lx >= 2 then
              let n = String.lowercase (String.trim (List.hd lx)) in
              let v = String.trim (List.hd (List.tl lx)) in
              let m = Dict.add n v m in
              rdHead st m ""
            else
              rdHead st m ""
      else if c = '\r' then
        rdHead st m line
      else
        rdHead st m (line ^ String.make 1 c)
  in
  let (st, m) = rdHead "" Dict.empty "" in
  let fp = { fp with status = st; resp_header = m } in
  (*let _ = print_endline fp.status in*)
  match fp.status with
  | "302" | "303" | "307" | "308" ->
    let _ = close fp in
    (*let _ = print_endline (resp_header fp "Location") in*)
    open_get req_header (resp_header fp "Location")
  | _ ->
    fp

let read_all_chunked fp =
  let eof = ref false in
  let ac = ref [] in
  let current = ref ([], 0, 0, [], 0) in
  let _ =
    while not !eof do
      match getc fp with
      | Some c ->
        let (ck, ckSize, n, data, dataSize) = !current in
        let ct = 
          if ck = [] then
            let ckSize = ckSize + 1 in
            ([c], ckSize, 0, [], 0)
          else
            if n = 0 then
              let cx = c :: ck in
              let ckSize = ckSize + 1 in
              if c = '\n' && List.hd ck = '\r' && ckSize > 2 then
                let hx = "0x" ^ (String.trim (CharList.to_string_rev cx)) in
                (*let _ = print_endline hx in*)
                let n = int_of_string hx in
                let _ = 
                  if n = 0 then
                    eof := true
                  else
                    ()
                in
                (cx, ckSize, n, data, dataSize)
              else
                (cx, ckSize, n, data, dataSize)
            else
              let data = c :: data in
              let dataSize = dataSize + 1 in
              if dataSize >= n then
                let _ = ac := data :: !ac in
                ([], 0, 0, [], 0)
              else
                (ck, ckSize, n, data, dataSize)
        in
        current := ct
      | _ ->
        eof := true
    done
  in
  let sx =
    List.map
      (fun d -> CharList.to_string_rev d)
      (List.rev !ac)
  in
  String.concat "" sx

let read_all_sub fp =
  if resp_header fp "Transfer-Encoding" = "chunked" then
    read_all_chunked fp
  else
    let cls = resp_header fp "Content-Length" in
    if not (String.is_empty cls) then
      read fp (int_of_string cls)
    else
      read fp (-1)

let read_all req_header url =
  let fp = open_get req_header url in
  let d = read_all_sub fp in
  let _ = close fp in
  d

let download url filePath ctypes =
  let _ = print_endline (url ^ " => " ^ filePath) in
  try
    let fp = open_get [] url in
    let ok =
      if List.is_empty ctypes then
        true
      else
        let ct = resp_header fp "Content-Type" in
        (*let _ = print_endline ct in*)
        let b = not (List.is_empty (List.filter (fun s -> String.find ct s <> -1) ctypes)) in
        let _ =
          if not b then
            print_endline ("NG Content-Type: " ^ ct)
          else
            ()
        in
        b
    in
    let result = 
      if ok then
        let cl = resp_header fp "Content-Length" in
        (*let _ = print_endline cl in*)
        let fileSize = if String.is_empty cl then 0 else int_of_string cl in
        if fileSize = 0 then
          let d = read_all_sub fp in
          let _ = Unix.write_file_all filePath d in
          true
        else
          let blockSize = 4096 in
          let oc = open_out_bin filePath in
          let rec loop rs dc =
            if rs = fileSize then
              rs
            else
              let rest = fileSize - rs in
              let reqSize = if rest > blockSize then blockSize else rest in
              let data = read fp reqSize in
              let sz = String.length data in
              if sz = 0 then
                (*let _ = print_endline "*" in*)
                rs
              else
                let rs = rs + sz in
                let rec progress dc = 
                  if 50 * rs / fileSize < dc then
                    dc
                  else
                    let _ = print_char '=' in
                    let _ = flush stdout in
                    progress (dc + 1)
                in
                let dc = 
                  if fileSize > 0 then
                    progress dc
                  else
                    dc
                in
                let _ = output oc data 0 (String.length data) in
                loop rs dc
                (*if sz < blockSize then
                  rs
                else
                  loop rs dc*)
          in
          let readSize = loop 0 0 in
          let _ = close_out oc in
          if fileSize > 0 then
            if fileSize <> readSize then
              let _ = Sys.remove filePath in
              let _ = print_endline " NG" in
              false
            else
              let _ = print_endline "" in
              true
          else
            true
        else
          false
    in
    let _ = close fp in
    result
  with
  | _ -> 
    let _ = print_endline "NG" in
    false
